import argparse
import os
import logging
from datetime import datetime

from track.trackextractor import TrackExtractor
from ml_tools.logs import init_logging
from config.config import Config
from config.classifyconfig import ModelConfig

import absl.logging


def str2bool(v):
    if isinstance(v, bool):
        return v
    if v.lower() in ("yes", "true", "t", "y", "1"):
        return True
    elif v.lower() in ("no", "false", "f", "n", "0"):
        return False
    else:
        raise argparse.ArgumentTypeError("Boolean value expected.")


def main(cmd_args=None):
    logging.root.removeHandler(absl.logging._absl_handler)
    absl.logging._warn_preinit_stderr = False
    parser = argparse.ArgumentParser()

    parser.add_argument(
        "source",
        help='a CPTV file to process, or a folder name, or "all" for all files within subdirectories of source folder.',
    )
    parser.add_argument(
        "-p",
        "--preview-type",
        help="Create MP4 previews of this type (can be slow), this overrides the config",
    )
    parser.add_argument(
        "-v", "--verbose", action="count", help="Display additional information."
    )
    parser.add_argument(
        "-o",
        "--meta-to-stdout",
        action="count",
        help="Print metadata to stdout instead of saving to file.",
    )
    parser.add_argument("-c", "--config-file", help="Path to config file to use")

    parser.add_argument(
        "-T", "--timestamps", action="store_true", help="Emit log timestamps"
    )

    parser.add_argument(
        "--retrack",
        type=str2bool,
        nargs="?",
        const=True,
        default=None,
        help="Use existing metadata to correct tracks",
    )
    parser.add_argument(
        "--cache",
        type=str2bool,
        nargs="?",
        const=True,
        default=None,
        help="Dont keep video frames in memory for classification later, but cache them to disk (Best for large videos, but slower)",
    )
    if cmd_args is None:
        args = parser.parse_args()
    else:
        args = parser.parse_args(cmd_args)
    config = Config.load_from_file(args.config_file)
    config.validate()
    init_logging(args.timestamps)

    # parse command line arguments
    if args.preview_type:
        config.classify.preview = args.preview_type

    if args.verbose:
        config.classify_tracking.verbose = True

    to_stdout = config.classify.meta_to_stdout
    if args.meta_to_stdout:
        config.classify.meta_to_stdout = True
    extractor = TrackExtractor(config, cache_to_disk=args.cache, retrack=args.retrack)
    extractor.extract(args.source, to_stdout=to_stdout)


if __name__ == "__main__":
    main()



"""
classifier-pipeline - this is a server side component that manipulates cptv
files and to create a classification model of animals present
Copyright (C) 2018, The Cacophony Project

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
"""

import attr
import os
import logging
from os import path
from .defaultconfig import DefaultConfig
from ml_tools.rectangle import Rectangle


@attr.s
class BuildConfig(DefaultConfig):
    test_clips_folder = attr.ib()
    banned_clips = attr.ib()
    segment_length = attr.ib()
    segment_spacing = attr.ib()
    segment_min_avg_mass = attr.ib()
    min_frame_mass = attr.ib()
    filter_by_lq = attr.ib()
    max_segments = attr.ib()
    thermal_diff_norm = attr.ib()
    tag_precedence = attr.ib()
    excluded_tags = attr.ib()
    country = attr.ib()
    use_segments = attr.ib()
    max_frames = attr.ib()

    EXCLUDED_TAGS = ["poor tracking", "part", "untagged", "unidentified"]
    NO_MIN_FRAMES = ["stoat", "mustelid", "weasel", "ferret"]
    # country bounding boxs
    COUNTRY_LOCATIONS = {
        "AU": Rectangle.from_ltrb(
            113.338953078, -10.6681857235, 153.569469029, -43.6345972634
        ),
        "NZ": Rectangle.from_ltrb(
            166.509144322, -34.4506617165, 178.517093541, -46.641235447
        ),
    }

    DEFAULT_GROUPS = {
        0: [
            "bird",
            "false-positive",
            "hedgehog",
            "possum",
            "rodent",
            "mustelid",
            "cat",
            "kiwi",
            "dog",
            "leporidae",
            "human",
            "insect",
            "pest",
        ],
        1: ["unidentified", "other"],
        2: ["part", "bad track"],
        3: ["default"],
    }

    @classmethod
    def load(cls, build):
        return cls(
            test_clips_folder=build["test_clips_folder"],
            banned_clips=load_banned_clips_file(build["banned_clips"]),
            segment_length=build["segment_length"],
            segment_spacing=build["segment_spacing"],
            segment_min_avg_mass=build["segment_min_avg_mass"],
            min_frame_mass=build["min_frame_mass"],
            filter_by_lq=build["filter_by_lq"],
            max_segments=build["max_segments"],
            thermal_diff_norm=build["thermal_diff_norm"],
            tag_precedence=build["tag_precedence"],
            excluded_tags=build["excluded_tags"],
            country=build["country"],
            use_segments=build["use_segments"],
            max_frames=build["max_frames"],
        )

    @classmethod
    def get_defaults(cls):
        return cls(
            test_clips_folder=None,
            banned_clips=None,
            segment_length=25,
            segment_spacing=1,
            segment_min_avg_mass=10,
            min_frame_mass=10,
            filter_by_lq=False,
            max_segments=3,
            thermal_diff_norm=False,
            tag_precedence=BuildConfig.DEFAULT_GROUPS,
            excluded_tags=BuildConfig.EXCLUDED_TAGS,
            country=None,
            use_segments=True,
            max_frames=75,
        )

    def validate(self):
        return True

    def test_clips(self):
        if not self.test_clips_folder or not path.exists(self.test_clips_folder):
            return None
        if not os.path.isdir(self.test_clips_folder):
            return None
        test_clips = []
        for f in os.listdir(self.test_clips_folder):
            file_path = path.join(self.test_clips_folder, f)
            if path.isfile(file_path) and path.splitext(file_path)[1] == ".list":
                with open(file_path) as stream:
                    for line in stream:
                        if line.strip() == "":
                            continue
                        try:
                            clips = line.split(",")
                            # print("line is", line, "clips are", clips)
                            for clip in clips:
                                test_clips.append(int(clip.strip()))
                        except:
                            logging.warn(
                                "Could not parse clip_id %s from %s",
                                line,
                                file_path,
                            )
        return test_clips


def load_banned_clips_file(filename):
    if not filename or not os.path.exists(filename):
        return None
    if not path.isfile(filename):
        return None
    files = []

    with open(filename) as stream:
        for line in stream:
            files.append(line.strip())
    return files



import numpy as np
from datetime import datetime
import logging
import os
import yaml
import multiprocessing
from datetime import timedelta
import time
import psutil
import cv2
from pathlib import Path

from ml_tools.mpeg_creator import MPEGCreator
from piclassifier.recorder import Recorder
from track.cliptrackextractor import ClipTrackExtractor
from ml_tools.logs import init_logging
from .eventreporter import log_event


VIDEO_EXT = ".mp4"

# FOURCC = cv2.VideoWriter_fourcc(*"avc1")
# JUST FOR TEST
# VIDEO_EXT = ".avi"
# FOURCC = cv2.VideoWriter_fourcc("M", "J", "P", "G")

# gp this should work on pi, but causing lots of headaches
# wont set bitrate, wont play back on ubuntu, should check again in future
# as is very fast
# CODEC = "h264_v4l2m2m"
CODEC = "libx264"
BITRATE = "1M"
PIX_FMT = "yuv420p"


class IRRecorder(Recorder):
    def __init__(
        self,
        thermal_config,
        headers,
        name="IRRecorder",
        **args,
    ):
        super().__init__(
            thermal_config,
            headers,
            name,
            VIDEO_EXT,
            *args,
        )
        self.res_x = headers.res_x
        self.res_y = headers.res_y
        self.fps = headers.fps

    def new_recording(self, background_frame, preview_frames, temp_thresh, frame_time):
        back = None
        if background_frame is not None and len(background_frame.shape) == 2:
            back = background_frame[:, :, np.newaxis]
            back = np.repeat(back, 3, axis=2)
        # preview_frames.insert(0, back)

        self.rec_p = multiprocessing.Process(
            target=record,
            args=(
                self.frame_q,
                self.filename,
                self.fps,
                back,
                preview_frames,
                self.name,
            ),
        )
        self.rec_p.start()
        return True

    def final_name(self):
        return self.output_dir / self.filename.name


#
# def new_temp_name(frame_time):
#     return datetime.fromtimestamp(frame_time).strftime("%Y%m%d-%H%M%S.%f" + VIDEO_EXT)


def write_frame(writer, frame):
    yuv = cv2.cvtColor(frame, cv2.COLOR_BGR2YUV_I420)
    writer.next_frame(yuv, frame.shape)


def record(queue, filename, fps, background=None, init_frames=None, name="IRRecorder"):
    init_logging()
    frames = 0
    try:
        logging.info("%s Recorder %s started", name, filename.resolve())

        writer = MPEGCreator(
            filename, fps=fps, codec=CODEC, bitrate=BITRATE, pix_fmt=PIX_FMT
        )
        if background is not None:
            write_frame(writer, background)
            frames += 1
        if init_frames is not None:
            for frame in init_frames:
                write_frame(writer, frame)
                frames += 1
        while True:
            frame = queue.get()
            if isinstance(frame, int) and frame == 0:
                writer.close()
                break
            write_frame(writer, frame)
            frames += 1

    except Exception as ex:
        logging.error("%s Error Recording %s", name, filename.resolve(), exc_info=True)
        log_event("error-recording", ex)
        try:
            writer.close()
        except:
            pass
    logging.info("%s Recorder %s Written %s", name, filename.resolve(), frames)



import h5py
import os
import numpy as np


from ml_tools.tools import get_clipped_flow
from ml_tools.frame import Frame, TrackChannels


class FrameCache:
    def __init__(self, cptv_name, keep_open=True, delete_if_exists=True):
        basename = os.path.splitext(cptv_name)[0]
        self.filename = basename + ".cache"
        self.db = None
        self.keep_open = keep_open
        self.num_farmes = 0
        if delete_if_exists:
            self.delete()

        f = h5py.File(self.filename, "w")
        f.create_group("frames")
        f.close()

    def add_frame(self, frame):
        self.open()
        frames = self.db["frames"]
        frame_group = frames.create_group(str(frame.frame_number))
        frame_group.attrs["ffc_affected"] = frame.ffc_affected
        height, width = frame.thermal.shape

        chunks = (1, height, width)
        channels = []
        dims = 0
        data = []
        if frame.thermal is not None:
            channels.append(TrackChannels.thermal)
            dims += 1
            data.append(np.float32(frame.thermal))
        if frame.filtered is not None:
            channels.append(TrackChannels.filtered)
            dims += 1
            data.append(np.float32(frame.filtered))

        if frame.flow is not None:
            channels.append(TrackChannels.flow)
            scaled_flow = get_clipped_flow(frame.flow)
            scaled_flow_h = np.float32(scaled_flow[:, :, 0])
            scaled_flow_v = np.float32(scaled_flow[:, :, 1])
            data.append(scaled_flow_h)
            data.append(scaled_flow_v)
            dims += 2
        if frame.mask is not None:
            channels.append(TrackChannels.mask)
            data.append(np.float32(frame.mask))
            dims += 1
        frame_group.attrs["channels"] = np.uint8(channels)

        dims = (dims, height, width)
        frame_node = frame_group.create_dataset(
            "frame", dims, chunks=chunks, dtype=np.float32
        )

        frame_node[:, :, :] = data
        if not self.keep_open:
            self.close()

    def get_frame(self, frame_number):
        self.open()
        frame = None
        if str(frame_number) in self.db["frames"]:
            frame_group = self.db["frames"][str(frame_number)]
            frame = frame_group["frame"]
            ffc_affected = frame_group.attrs["ffc_affected"]
            channels = frame_group.attrs["channels"]
            frame = Frame.from_channels(
                frame,
                channels,
                frame_number,
                flow_clipped=True,
                ffc_affected=ffc_affected,
            )

        if not self.keep_open:
            self.close()
        return frame

    def close(self):
        if self.db:
            self.db.close()
            self.db = None

    def open(self, mode="a"):
        if not self.db:
            self.db = h5py.File(self.filename, mode)

    def delete(self):
        if self.db:
            self.close()
        if os.path.exists(self.filename):
            os.remove(self.filename)



import attr
import numpy as np


@attr.s(eq=False)
class Rectangle:
    """Defines a rectangle by the topleft point and width / height."""

    x = attr.ib()
    y = attr.ib()
    width = attr.ib()
    height = attr.ib()

    @staticmethod
    def from_ltrb(left, top, right, bottom):
        """Construct a rectangle from left, top, right, bottom co-ords."""
        return Rectangle(left, top, width=right - left, height=bottom - top)

    def to_ltrb(self):
        """Return rectangle as left, top, right, bottom co-ords."""
        return [self.left, self.top, self.right, self.bottom]

    def to_ltwh(self):
        """Return rectangle as left, top, right, bottom co-ords."""
        return [self.left, self.top, self.width, self.height]

    def copy(self):
        return Rectangle(self.x, self.y, self.width, self.height)

    @property
    def mid(self):
        return (self.mid_x, self.mid_y)

    @property
    def mid_x(self):
        return self.x + self.width / 2

    @property
    def mid_y(self):
        return self.y + self.height / 2

    @property
    def left(self):
        return self.x

    @property
    def top(self):
        return self.y

    @property
    def right(self):
        return self.x + self.width

    @property
    def bottom(self):
        return self.y + self.height

    @left.setter
    def left(self, value):
        old_right = self.right
        self.x = value
        self.right = old_right

    @top.setter
    def top(self, value):
        old_bottom = self.bottom
        self.y = value
        self.bottom = old_bottom

    @right.setter
    def right(self, value):
        self.width = value - self.x

    @bottom.setter
    def bottom(self, value):
        self.height = value - self.y

    def overlap_area(self, other):
        """Compute the area overlap between this rectangle and another."""
        x_overlap = max(0, min(self.right, other.right) - max(self.left, other.left))
        y_overlap = max(0, min(self.bottom, other.bottom) - max(self.top, other.top))
        return x_overlap * y_overlap

    def crop(self, bounds):
        """Crops this rectangle so that it fits within given bounds"""
        self.left = min(bounds.right, max(self.left, bounds.left))
        self.top = min(bounds.bottom, max(self.top, bounds.top))
        self.right = max(bounds.left, min(self.right, bounds.right))
        self.bottom = max(bounds.top, min(self.bottom, bounds.bottom))

    def subimage(self, image):
        """Returns a subsection of the original image bounded by this rectangle
        :param image mumpy array of dims [height, width]
        """
        return image[
            self.top : self.top + self.height, self.left : self.left + self.width
        ]

    def enlarge(self, border, max=None):
        """Enlarges this by border amount in each dimension such that it fits
        within the boundaries of max"""
        self.left -= border
        self.right += border
        self.top -= border
        self.bottom += border
        if max:
            self.crop(max)

    def contains(self, x, y):
        """Is this point contained in the rectangle"""
        return self.left <= x and self.right >= x and self.top >= y and self.bottom <= y

    @property
    def area(self):
        return int(self.width) * self.height

    def __repr__(self):
        return "(x{0},y{1},x2{2},y2{3})".format(
            self.left, self.top, self.right, self.bottom
        )

    def __str__(self):
        return "<(x{0},y{1})-h{2}xw{3}>".format(self.x, self.y, self.height, self.width)

    def meta_dictionary(self):
        # Return object as dictionary without is_along_border,was_cropped and id for saving to json
        region_info = attr.asdict(
            self,
            filter=lambda attr, value: attr.name
            not in ["is_along_border", "was_cropped", "id", "centroid"],
        )
        # region_info["centroid"][0] = round(region_info["centroid"][0], 1)
        # region_info["centroid"][1] = round(region_info["centroid"][1], 1)
        if region_info["pixel_variance"] is not None:
            region_info["pixel_variance"] = round(region_info["pixel_variance"], 2)
        else:
            region_info["pixel_variance"] = 0
        return region_info



#!/usr/bin/python3
import argparse

import socket
from cptv import CPTVReader
import sys
import numpy as np
import h5py

SOCKET_NAME = "/var/run/lepton-frames"
test_cptv = "test.cptv"
test_h5py = "/home/zaza/Cacophony/classifier-pipeline/pithermal65.h5py"


def parse_args():
    parser = argparse.ArgumentParser()

    parser.add_argument("--cptv", help="a CPTV file to send", default="test.cptv")
    parser.add_argument("--h5", help="a h5py to send")
    parser.add_argument("-clip_id", help="Clip id of h5py file to send")
    args = parser.parse_args()
    return args


def send_cptv(filename, socket):
    """
    Loads a cptv file, and prepares for track extraction.
    """
    with open(filename, "rb") as f:
        reader = CPTVReader(f)
        for i, frame in enumerate(reader):
            f = np.uint16(frame.pix).byteswap()
            socket.sendall(f)
            print("sending frame {}".format(i))


def send_h5py(filename, clip_id, socket):
    db = h5py.File(filename, mode="r")
    clips = db["clips"]
    if clip_id:
        clip = clips[str(clip_id)]
        send_clip(clip, socket)
    else:
        for clip_id in clips:
            send_clip(clips[clip_id], socket)
    db.close()


def send_clip(clip, socket):
    frames = clip["frames"]
    frame_ids = []
    for frame_id in frames:
        frame_ids.append(int(frame_id))

    frame_ids.sort()
    for i in range(1):
        for frame_id in frame_ids:
            f = np.uint16(frames[str(frame_id)]).byteswap()
            socket.sendall(f)


def main():
    args = parse_args()
    sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)

    try:
        sock.connect(SOCKET_NAME)
        if args.h5:
            print("sending h5 {} clip {}".format(args.h5, args.clip_id))
            send_h5py(args.h5, args.clip_id, sock)
        else:
            print("sending cptv {}".format(args.cptv))
            send_cptv(args.cptv, sock)
    except socket.error as msg:
        print(msg)
        sys.exit(1)
    finally:
        # Clean up the connection
        sock.close()


if __name__ == "__main__":
    main()



"""A window time frame, which can be relative to sunset and sunrise"""

from datetime import datetime, timedelta, time
import logging
import enum
from astral import Location

from ml_tools import tools


class WindowStatus(enum.Enum):
    """Types of frames"""

    before = 0
    inside = 1
    after = 2
    non_stop = 3


class TimeWindow:
    def __init__(self, start, end, lat, lng, altitude):
        self.start = start
        self.end = end

        if self.start.any_time or self.end.any_time:
            assert self.start.any_time and self.end.any_time
        self.location = None
        self.last_sunrise_check = None
        self.non_stop = not self.use_sunrise_sunset() and self.start.dt == self.end.dt
        if lat is not None and lng is not None:
            self.set_location(lat, lng, altitude)
        # ensure starting window is valid
        # set end or start to following day if needed

        if not self.end.is_relative and self.end.is_after():

            self.end.dt = self.end.dt + timedelta(days=1)
        # set start to previous day if needed

        if (
            not self.start.is_relative
            and self.start.is_after()
            and datetime.now().time() > self.end.dt.time()
        ):
            self.start.dt = self.start.dt + timedelta(days=1)
        if (
            self.start.dt
            and not self.start.is_relative
            and self.end.is_before()
            and self.start.is_before()
            and self.start.dt > self.end.dt
        ):

            self.start.dt = self.start.dt + timedelta(days=-1)

        assert self.start.dt <= self.end.dt

    def clone(self):
        new_window = TimeWindow(
            self.start.clone(),
            self.end.clone(),
            self.location.latitude,
            self.location.longitude,
            self.location.altitude,
        )
        new_window.location = self.location
        new_window.last_sunrise_check = self.last_sunrise_check
        new_window.non_stop = self.non_stop
        return new_window

    def next_start(self):
        return next_time(self.start)

    def next_end(self):
        return next_time(self.end)

    def use_sunrise_sunset(self):
        return self.start.is_relative or self.end.is_relative

    def window_status(self):
        if self.non_stop:
            return WindowStatus.non_stop
        if self.use_sunrise_sunset():
            self.update_sun_times()
        if self.start.is_before():
            return WindowStatus.before
        if self.end.is_before():
            return WindowStatus.inside
        return WindowStatus.after

    def next_window(self):
        if self.use_sunrise_sunset():
            # update to tomorrows times
            self.update_sun_times(True)
        # handle case that only one is relative
        if self.start.dt is not None and not self.start.is_relative:
            self.start.dt = self.start.dt + timedelta(days=1)
        if self.end.dt is not None and not self.end.is_relative:
            self.end.dt = self.end.dt + timedelta(days=1)
        logging.info(
            "Updated to next window start %s end %s", self.start.dt, self.end.dt
        )

    def inside_window(self):
        if self.start.time == self.end.time:
            return True
        if self.start.is_after() and self.end.is_after():
            self.next_window()
        return self.start.is_after() and self.end.is_before()

    def update_sun_times(self, next_window=False):
        if not self.use_sunrise_sunset():
            return

        if self.location is None:
            raise ValueError(
                "Location must be set for relative times, by calling set_location"
            )
        # if next_window:
        # date = date + timedelta(days=1)
        # after_window = self.start.is_before() or self.end.is_before()
        # if self.last_sunrise_check is not None and (
        #     self.start.is_before() or self.end.is_before()
        # ):
        if (
            self.last_sunrise_check is None
            or next_window
            or datetime.now() > self.end.dt
        ):
            date = datetime.now().date()
            if self.last_sunrise_check is not None and next_window:
                date = self.last_sunrise_check + timedelta(days=1)
            elif next_window:
                date = date + timedelta(days=1)
            sun_times = self.location.sun(date=date)
            self.last_sunrise_check = date
            if self.start.is_relative:
                self.start.dt = sun_times["sunset"] + timedelta(
                    seconds=self.start.offset_s
                )
                self.start.dt = self.start.dt.replace(tzinfo=None)

            if self.end.is_relative:
                self.end.dt = sun_times["sunrise"] + timedelta(
                    seconds=self.end.offset_s
                )
                self.end.dt = self.end.dt.replace(tzinfo=None)
                if datetime.now() > self.end.dt:
                    date = date + timedelta(days=1)
                    sun_times = self.location.sun(date=date)
                    self.end.dt = sun_times["sunrise"] + timedelta(
                        seconds=self.end.offset_s
                    )
                    self.end.dt = self.end.dt.replace(tzinfo=None)
            if (
                not next_window == True
                and self.start.is_relative
                and self.start.is_before()
                and self.end.is_before()
            ):
                date = date + timedelta(days=-1)
                sun_times = self.location.sun(date=date)
                self.start.dt = sun_times["sunset"] + timedelta(
                    seconds=self.start.offset_s
                )
                self.start.dt = self.start.dt.replace(tzinfo=None)

            logging.info(
                "Updated sun times start is {} end is {}".format(
                    self.start.dt, self.end.dt
                )
            )

    def set_location(self, lat, lng, altitude=0):
        self.location = Location()
        self.location.latitude = lat
        self.location.longitude = lng
        self.location.altitude = altitude
        self.location.timezone = tools.get_timezone_str(lat, lng)

        self.update_sun_times()


class RelAbsTime:
    def __init__(self, time_str, default_offset=None, default_time=None):
        self.is_relative = False
        self.offset_s = None
        self.dt = None
        self.any_time = False
        if time_str == "" or (
            time_str is None and default_offset is None and default_time is None
        ):
            self.dt = datetime.combine(
                datetime.now(),
                time(12, 0),
            )
            self.any_time = True
            return

        try:
            self.dt = datetime.combine(
                datetime.now(), datetime.strptime(time_str, "%H:%M").time()
            )
        except (ValueError, TypeError):
            if not time_str:
                self.offset_s = default_offset
            elif isinstance(time_str, int) or time_str.isnumeric():
                self.offset_s = int(time_str)
            else:
                self.offset_s = self.parse_duration(time_str, default_offset)

            if self.offset_s is None and default_time:
                self.dt = default_time
            else:
                self.is_relative = True

    def clone(self):
        new_time = RelAbsTime("")
        new_time.is_relative = self.is_relative
        new_time.offset_s = self.offset_s
        new_time.dt = self.dt
        new_time.any_time = self.any_time
        return new_time

    @property
    def time(self):
        return self.dt.time() if self.dt is not None else None

    def is_after(self, ignore_date=False):
        if ignore_date:
            return self.any_time or datetime.now().time() > self.dt.time()
        return self.any_time or datetime.now() > self.dt

    def is_before(self, ignore_date=False):
        if ignore_date:
            return self.any_time or datetime.now().time() < self.dt.time()
        return self.any_time or datetime.now() < self.dt

    def parse_duration(self, time_str, default_offset=None):
        if not time_str:
            return default_offset

        time_str = time_str.strip()
        unit = time_str[-1]
        if unit.isalpha():
            try:
                offset = float(time_str[:-1])
            except ValueError:
                return default_offset
            if unit == "s":
                return offset
            elif unit == "m":
                return offset * 60
            elif unit == "h":
                return offset * 60 * 60
            return offset

        try:
            offset = float(time_str)
            return offset * 60
        except ValueError:
            pass
        return default_offset


def next_time(rel_time):
    if rel_time.any_time:
        return None
    else:
        return rel_time.dt



"""
HeaderInfo describes a thermal cameras specs.
When a thermal camera first connects to the socket it will send some header
information describing it's specs e.g. Resolution, Frame rate
"""

import yaml
import attr


@attr.s
class HeaderInfo:
    X_RESOLUTION = "ResX"
    Y_RESOLUTION = "ResY"
    FPS = "FPS"
    MODEL = "Model"
    BRAND = "Brand"
    PIXEL_BITS = "PixelBits"
    FRAME_SIZE = "FrameSize"
    SERIAL = "CameraSerial"
    FIRMWARE = "Firmware"

    res_x = attr.ib()
    res_y = attr.ib()
    fps = attr.ib()
    brand = attr.ib()
    model = attr.ib()
    frame_size = attr.ib()
    pixel_bits = attr.ib()
    serial = attr.ib()
    firmware = attr.ib()

    @classmethod
    def parse_header(cls, raw_string):
        raw = yaml.safe_load(raw_string)
        headers = cls(
            res_x=raw.get(HeaderInfo.X_RESOLUTION),
            res_y=raw.get(HeaderInfo.Y_RESOLUTION),
            fps=raw.get(HeaderInfo.FPS),
            brand=raw.get(HeaderInfo.BRAND),
            model=raw.get(HeaderInfo.MODEL),
            serial=raw.get(HeaderInfo.SERIAL),
            frame_size=raw.get(HeaderInfo.FRAME_SIZE),
            pixel_bits=raw.get(HeaderInfo.PIXEL_BITS),
            firmware=raw.get(HeaderInfo.FIRMWARE),
        )
        headers.firmware = str(headers.firmware)
        if headers.res_x and headers.res_y:
            if not headers.pixel_bits and headers.frame_size:
                headers.pixel_bits = int(
                    8 * headers.frame_size / (headers.res_x * headers.res_y)
                )
            elif not headers.frame_size and headers.pixel_bits:
                headers.frame_size = int(
                    headers.res_x * headers.res_y * headers.pixel_bits / 8
                )
        headers.validate()

        return headers

    def validate(self):
        if not (self.res_x and self.res_y and self.fps and self.pixel_bits):
            raise ValueError(
                "header info is missing a required field ({}, {}, {} and/or {})".format(
                    HeaderInfo.X_RESOLUTION,
                    HeaderInfo.Y_RESOLUTION,
                    HeaderInfo.FPS,
                    HeaderInfo.PIXEL_BITS,
                )
            )
        return True

    def as_dict(self):
        return attr.asdict(self)



"""

Author Matthew Aitchison

Date December 2017

Handles reading and writing tracks (or segments) to a large database.  Uses HDF5 as a backing store.

"""

import h5py
import os
import logging
import filelock
import numpy as np
from dateutil.parser import parse as parse_date
from .frame import Frame, TrackChannels
import json

import numpy as np
from track.region import Region
from ml_tools.datasetstructures import TrackHeader, ClipHeader

special_datasets = [
    "tag_frames",
    "original_frames",
    "background_frame",
    "predictions",
    "overlay",
]


class HDF5Manager:
    """Class to handle locking of HDF5 files."""

    LOCK_FILE = "/var/lock/classifier-hdf5.lock"
    READ_ONLY = False

    def __init__(self, db, mode="r"):
        self.mode = mode
        self.f = None
        self.db = db
        self.lock = filelock.FileLock(HDF5Manager.LOCK_FILE, timeout=60 * 3)
        filelock.logger().setLevel(logging.ERROR)

    def __enter__(self):
        # note: we might not have to lock when in read only mode?
        # this could improve performance
        if HDF5Manager.READ_ONLY and self.mode != "r":
            raise ValueError("Only read can be done in readonly mode")
        if not HDF5Manager.READ_ONLY:
            self.lock.acquire()
        self.f = h5py.File(self.db, self.mode)
        return self.f

    def __exit__(self, exc_type, exc_val, exc_tb):
        try:
            self.f.close()
        finally:
            if not HDF5Manager.READ_ONLY:
                self.lock.release()


class TrackDatabase:
    def __init__(self, database_filename, read_only=False):
        """
        Initialises given database.  If database does not exist an empty one is created.
        :param database_filename: filename of database
        """

        self.database = database_filename
        # if not os.path.exists(database_filename):
        #     logging.info("Creating new database %s", database_filename)
        #     f = h5py.File(database_filename, "w")
        #     f.create_group("clips")
        #     f.close()
        HDF5Manager.READ_ONLY = read_only

    def set_read_only(self, read_only):
        HDF5Manager.READ_ONLY = read_only

    def has_clip(self, clip_id):
        """
        Returns if database contains track information for given clip
        :param clip_id: name of clip
        :return: If the database contains given clip
        """
        with HDF5Manager(self.database) as f:
            clips = f["clips"]
            has_record = clip_id in clips and "finished" in clips[clip_id].attrs
            if has_record:
                return True
        return False

    def has_prediction(self, clip_id):
        with HDF5Manager(self.database, "a") as f:
            clips = f["clips"]
            # has_record = clip_id in clips and "finished" in clips[clip_id].attrs
            clip = clips[clip_id]
            # if has_record:
            return clip.attrs.get("has_prediction", False)
        return False

    #
    # def add_prediction_data(self, clip_id, track, predictions, labels=None):
    #     """
    #     Add prediction data as a dataset to the track
    #     data should be  an array of int16 array
    #     """
    #     logging.info("Add pre data %s", predictions)
    #     track_attrs = track.attrs
    #     prediction = predictions[0]
    #     predicted_tag = prediction.get("label")
    #     score = prediction.get("confidence")
    #     print("tag is", predicted_tag), prediction
    #     if predicted_tag is not None:
    #         track_attrs["correct_prediction"] = track_attrs.get("tag") == predicted_tag
    #         track_attrs["predicted"] = predicted_tag
    #         print("predicted", predicted_tag)
    #     track_attrs["predicted_confidence"] = int(round(100 * score))
    #
    #     pred_data = track.create_dataset(
    #         "predictions",
    #         predictions.shape,
    #         chunks=predictions.shape,
    #         dtype=predictions.dtype,
    #     )
    #     pred_data[:, :] = predictions
    #     if labels is not None:
    #         track_attrs["prediction_classes"] = labels

    def finished_processing(self, clip_id):
        with HDF5Manager(self.database, "a") as f:
            clip_node = f["clips"][clip_id]
            clip_node.attrs["finished"] = True

    def get_labels(self):
        with HDF5Manager(self.database) as f:
            return f.attrs.get("labels", None)

    def create_clip(self, clip, overwrite=True):
        """
        Creates a clip entry in database.
        :param clip_id: id of the clip
        :param tracker: if provided stats from tracker are used for the clip stats
        :param overwrite: Overwrites existing clip (if it exists).
        """
        logging.info("creating clip %s", clip.get_id())
        clip_id = str(clip.get_id())
        with HDF5Manager(self.database, "a") as f:
            clips = f["clips"]
            if overwrite and clip_id in clips:
                del clips[clip_id]
            group = clips.create_group(clip_id)
            if clip is not None:
                if clip.background is not None:
                    height, width = clip.background.shape
                    background_frame = group.create_dataset(
                        "background_frame",
                        (height, width),
                        chunks=(height, width),
                        dtype=clip.background.dtype,
                    )
                    background_frame[:, :] = clip.background
                group_attrs = group.attrs

                # group_attrs.update(clip.stats)
                group_attrs["filename"] = clip.source_file
                group_attrs["start_time"] = clip.video_start_time.isoformat()
                group_attrs["background_thresh"] = clip.background_thresh

                if clip.res_x and clip.res_y:
                    group_attrs["res_x"] = clip.res_x
                    group_attrs["res_y"] = clip.res_y
                if clip.crop_rectangle:
                    group_attrs["edge_pixels"] = clip.crop_rectangle.left

                group_attrs["mean_background_value"] = clip.stats.mean_background_value
                group_attrs["threshold"] = clip.stats.threshold
                group_attrs["max_temp"] = clip.stats.max_temp
                group_attrs["min_temp"] = clip.stats.min_temp
                group_attrs["mean_temp"] = clip.stats.mean_temp
                group_attrs["filtered_deviation"] = clip.stats.filtered_deviation
                group_attrs["filtered_sum"] = clip.stats.filtered_sum
                group_attrs["temp_thresh"] = clip.stats.temp_thresh

                group_attrs["frame_temp_min"] = clip.stats.frame_stats_min
                group_attrs["frame_temp_max"] = clip.stats.frame_stats_max
                group_attrs["frame_temp_median"] = clip.stats.frame_stats_median
                group_attrs["frame_temp_mean"] = clip.stats.frame_stats_mean

                if clip.device:
                    group_attrs["device"] = clip.device
                group_attrs["frames_per_second"] = clip.frames_per_second
                if clip.location:
                    if "coordinates" in clip.location:
                        loc = clip.location["coordinates"]
                    else:
                        loc = (clip.location["lat"], clip.location["lng"])

                    group_attrs["location"] = loc
                if clip.station_id:
                    group_attrs["station_id"] = clip.station_id
                if clip.tags:
                    clip_tags = []
                    for track in clip.tags:
                        if track["what"]:
                            clip_tags.append(track["what"])
                        elif track["detail"]:
                            clip_tags.append(track["detail"])
                    group_attrs["tags"] = clip_tags
                group_attrs["ffc_frames"] = clip.ffc_frames

            f.flush()

    def latest_date(self):
        start_time = None

        with HDF5Manager(self.database) as f:
            clips = f["clips"]
            results = {}
            for clip_id in clips:
                clip_start = clips[clip_id].attrs["start_time"]
                if clip_start is not None:
                    clip_start = parse_date(clip_start)
                    if start_time is None or clip_start > start_time:
                        start_time = clip_start

        return start_time

    def get_all_clip_ids(self, before_date=None, after_date=None, label=None):
        """
        Returns a list of clip_id, track_id pairs.
        """
        with HDF5Manager(self.database) as f:
            clips = f["clips"]
            results = {}
            for clip_id in clips:
                clip = clips[clip_id]
                if not clip.attrs.get("finished"):
                    continue
                date = parse_date(clip.attrs["start_time"])
                if before_date and date >= before_date:
                    continue
                if after_date and date < after_date:
                    continue
                if label is not None:
                    if clip.attrs.get("tag") != label:
                        continue
                results[clip_id] = [track_id for track_id in clip]
        return results

    def get_clip_tracks_ids(self, clip_id):
        """
        Returns a list of clip_id, track_id pairs.
        """
        with HDF5Manager(self.database) as f:
            clips = f["clips"]
            tracks = []
            clip = clips[clip_id]
            if not clip.attrs.get("finished"):
                return tracks
            for track in clip:
                if track not in special_datasets:
                    tracks.append(track)
        return tracks

    def get_all_track_ids(self, before_date=None, after_date=None, label=None):
        """
        Returns a list of clip_id, track_id pairs.
        """
        with HDF5Manager(self.database) as f:
            clips = f["clips"]
            result = []
            for clip_id in clips:
                clip = clips[clip_id]
                if not clip.attrs.get("finished"):
                    continue
                date = parse_date(clip.attrs["start_time"])
                if before_date and date >= before_date:
                    continue
                if after_date and date < after_date:
                    continue
                for track in clip:
                    if label is not None:
                        if label != clip[track].attrs.get("tag"):
                            continue
                    if track not in special_datasets:
                        result.append((clip_id, track))
        return result

    def remove_tag_info(self, clip_id, track_id):
        with HDF5Manager(self.database, "a") as f:
            clip = f["clips"][clip_id]
            try:
                del clip["tag_frames"]
            except:
                pass

            if "tag" in clip.attrs:
                del clip.attrs["tag"]

            track = clip[track_id]
            track.attrs["tag_confirmed"] = False

    def update_tag(self, clip_id, track_id, frames, clip_tag, track_tag, regions=None):
        logging.info(
            "Tagging clip %s as %s and track %s as %s",
            clip_id,
            clip_tag,
            track_id,
            track_tag,
        )
        with HDF5Manager(self.database, "a") as f:
            clip = f["clips"][clip_id]
            clip_tags = clip.get("tag_frames")
            if clip_tags is None:
                clip_tags = clip.create_group("tag_frames")
                print("creating clip tags")
            tag_regions = clip_tags.get("tag_regions")
            if tag_regions is None:
                tag_regions = clip_tags.create_group("tag_regions")
            print("clip tag", clip_tag)
            if clip_tag in clip_tags.attrs:
                clip_frames = list(clip_tags.attrs[clip_tag])
            else:
                clip_frames = set()

            clip_frames = set(clip_frames)
            add_regions = clip_tag == track_tag
            label_regions = []
            if add_regions:
                if clip_tag in tag_regions.attrs:
                    label_regions = list(tag_regions.attrs[clip_tag])
                    for i in range(len(label_regions)):
                        r = label_regions[i]
                        r[6] = 1 if r[6] else 0
                        r = np.uint16(r)
                        label_regions[i] = r
            for f in frames:
                if f.frame_number not in clip_frames:
                    clip_frames.add(f.frame_number)
                if add_regions:
                    if f.region.blank:
                        continue
                    r_array = f.region.to_array()

                    unique_region = True
                    for other in label_regions:
                        diff = False
                        for i in range(len(other)):
                            if other[i] != r_array[i]:
                                diff = True
                                break
                        if not diff:
                            unique_region = False
                            break
                    if unique_region:
                        label_regions.append(r_array)

            if add_regions:
                tag_regions.attrs[clip_tag] = label_regions
                # for k, v in tag_regions.attrs.items():
                #     print("K", k, v)

            clip_frames = list(clip_frames)
            clip_frames.sort()
            clip_tags.attrs[clip_tag] = np.array(clip_frames)

            # clip.attrs["tag_frames"] = clip_tags
            clip.attrs["tag"] = clip_tag
            track = clip[track_id]
            if track_tag is not None:
                track.attrs["tag"] = track_tag
                track.attrs["tag_confirmed"] = True

    def get_track_meta(self, clip_id, track_id):
        """
        Gets metadata for given track
        :param clip_id:
        :param track_id:
        :return:
        """
        with HDF5Manager(self.database) as f:
            dataset = f["clips"][str(clip_id)][str(track_id)]
            result = self.dataset_track(dataset, track_id)

        return result

    def dataset_track(self, dataset, clip_id, track_id):
        result = hdf5_attributes_dictionary(dataset)
        regions = dataset.get("regions")[:]
        regions_by_frame = {}
        for r in regions:
            r = Region.region_from_array(r)
            regions_by_frame[r.frame_number] = r
        header = TrackHeader(
            clip_id=clip_id,
            track_id=int(track_id),
            label=result.get("human_tag"),
            num_frames=len(regions),
            regions=regions_by_frame,
            start_frame=result["start_frame"],
            confidence=result.get("human_tag_confidence"),
            human_tags=result.get("human_tags"),
            source_file=self.database,
            # frame_temp_median=frame_temp_median,
        )
        return header

    def get_track_predictions(self, clip_id, track_id):
        """
        Gets metadata for given track
        :param clip_id:
        :param track_id:
        :return:
        """
        with HDF5Manager(self.database) as f:
            track = f["clips"][clip_id][str(track_id)]
            if "predictions" in track:
                return track["predictions"][:]
        return None

    def get_clip_background(self):
        with HDF5Manager(self.database) as clip:
            frames = clip["frames"]
            if "background" in frames:
                return frames["background"][:]
        return None

    def get_clip_meta(self, clip_id=None):
        """
        Gets metadata for given clip
        :param clip_id:
        :return:
        """

        with HDF5Manager(self.database) as f:
            if "clips" in f:
                dataset = f["clips"][str(clip_id)]
            else:
                dataset = f
            result = hdf5_attributes_dictionary(dataset)
            clip_header = ClipHeader(
                clip_id=int(result["clip_id"]),
                station_id=result.get("station_id"),
                source_file=self.database,
                location=result.get("station_id"),
                camera=result.get("device_id"),
                frames_per_second=9,
                events=result.get("event", ""),
                trap=result.get("trap", ""),
                tracks=[],
                ffc_frames=result.get("ffc_frames"),
                rec_time=None,
                frame_temp_median=result.get("frame_temp_median"),
            )
        return clip_header

    def get_clip_tracks(self, clip_id=None):
        """
        Gets metadata for given clip
        :param clip_id:
        :return:
        """
        tracks = []
        clip_header = self.get_clip_meta()
        with HDF5Manager(self.database) as f:
            if "clips" in f:
                clip = f["clips"][str(clip_id)]
            else:
                clip = f["tracks"]
            for track_id in clip:
                if track_id in special_datasets:
                    continue
                track = self.dataset_track(
                    clip[track_id], clip_header.clip_id, track_id
                )
                track.station_id = clip_header.station_id
                clip_header.tracks.append(track)
        return clip_header

    def get_tag(self, clip_id, track_id):
        with HDF5Manager(self.database) as f:
            clips = f["clips"]
            track_node = clips[str(clip_id)][str(track_id)]
            return track_node.attrs["tag"]

    def get_frame(self, clip_id, track_id, frame, original=False):
        frames = self.get_track(
            clip_id, track_id, frame_numbers=[frame], original=original
        )
        if len(frames) == 1:
            return frames[0]
        return None

    def get_frames(
        self,
        frame_numbers=None,
        channels=None,
    ):
        # frames = []
        with HDF5Manager(self.database) as clip_node:
            raw_frames = clip_node["frames"]["thermals"][:]
            return raw_frames
        #     for frame_number, frame in enumerate(raw_frames):
        #         frames.append(
        #             Frame.from_channels([frame], [TrackChannels.thermal], frame_number)
        #         )
        # return frames

    def get_track(
        self,
        clip_id,
        track_id,
        start_frame=None,
        end_frame=None,
        frame_numbers=None,
        channels=None,
        crop=False,
    ):
        """
        Fetches a track data from database with optional slicing.
        :param clip_id: id of the clip
        :param track_id: id of the track
        :param start_frame: first frame of slice to return (inclusive).
        :param end_frame: last frame of slice to return (exclusive).
        :return: a list of numpy arrays of shape [channels, height, width] and of type np.int16
        """
        with HDF5Manager(self.database) as clip_node:
            tracks = clip_node["tracks"]
            track_node = tracks[str(track_id)]

            bounds = track_node["regions"]
            track_start = int(round(track_node.attrs.get("start_frame")))
            if start_frame is None:
                start_frame = track_start
            if end_frame is None:
                end_frame = int(round(track_node.attrs["end_frame"])) + 1
            bad_frames = track_node.attrs.get("skipepd_frames", [])
            result = []

            if frame_numbers is None:
                frame_iter = range(start_frame, end_frame)
            else:
                frame_iter = iter(frame_numbers)
            frames = clip_node["frames"]["thermals"][:]
            for frame_number in frame_iter:
                region = Region.region_from_array(bounds[frame_number - track_start])
                # assert region.frame_number == frame_number
                region.frame_number = frame_number
                frame = frames[frame_number]
                frame = region.subimage(frame)
                result.append(
                    Frame.from_channels(
                        [frame],
                        [TrackChannels.thermal],
                        frame_number,
                        region=region,
                    )
                )

        return result

    def remove_track(self, clip_id, track_id):
        with HDF5Manager(self.database, "a") as f:
            clips = f["clips"]
            clip_s = str(clip_id)
            track_s = str(track_id)
            if clip_s in clips:
                clip = clips[clip_s]
                if str(track_s) in clip:
                    del clip[track_s]
                return True
            else:
                return False

    def remove_clip(self, clip_id):
        """
        Deletes clip from database.
        Note, as per hdf5 the space will not be recovered.  If many files are deleted repacking the dataset might
        be a good idea.
        :param clip_id: id of clip to remove
        :returns: true if clip was deleted, false if it could not be found.
        """
        with HDF5Manager(self.database, "a") as f:
            clips = f["clips"]
            if clip_id in clips:
                del clips[clip_id]
                return True
            else:
                return False

    def set_sample_frames(self, clip_id, track_id, sample_frames):
        sample_frames.sort()
        with HDF5Manager(self.database, "a") as f:
            clips = f["clips"]
            clip_node = clips[str(clip_id)]
            track_node = clip_node[str(track_id)]
            track_node.attrs["sample_frames"] = np.uint16(sample_frames)

    def add_prediction(self, clip_id, track_id, track_prediction):
        logging.warn("Not adding prediction data as code needs to be written")

    # TODO IF NEEDED
    #     with HDF5Manager(self.database, "a") as f:
    #         clip = f["clips"][(str(clip_id))]
    #         track_node = clip[str(track_id)]
    #         predicted_tag = track_prediction.predicted_tag()
    #         if track_prediction.num_frames_classified > 0:
    #             self.all_class_confidences = track_prediction.class_confidences()
    #             predictions = np.int16(
    #                 np.around(100 * np.array(track_prediction.predictions))
    #             )
    #             predicted_confidence = int(round(100 * track_prediction.max_score))
    #
    #             self.add_prediction_data(
    #                 track_node,
    #                 predictions,
    #             )
    #         clip.attrs["has_prediction"] = True

    def add_prediction_data(self, track, model_predictions):
        """
        Add prediction data as a dataset to the track
        data should be  an array of int16 array
        """
        track_attrs = track.attrs

        model_group = track.create_group("model_predictions")
        for prediction in model_predictions:
            key = f'{prediction.get("model_id", 0)}'
            if key in model_group:
                continue
            pred_g = model_group.create_group(key)
            predicted_tag = prediction.get("label")
            if predicted_tag is not None:
                pred_g.attrs["correct_prediction"] = track_attrs["tag"] == predicted_tag
                pred_g.attrs["predicted"] = predicted_tag
                pred_g.attrs["confidence"] = prediction.get("confidence", 0)
            # track_attrs["predicted_confidence"] = prediction.get("confidence", 0)

            prediction_data = np.int16(prediction.get("predictions"))
            raw_predictions = pred_g.create_dataset(
                "predictions",
                prediction_data.shape,
                chunks=prediction_data.shape,
                dtype=prediction_data.dtype,
            )
            raw_predictions[:, :] = prediction_data
            labels = prediction.get("labels")
            if labels is not None:
                pred_g.attrs["prediction_classes"] = labels
            track_attrs["has_prediction"] = True

    def get_unique_clip_id(self):
        clip_ids = list(self.get_all_clip_ids().keys())
        clip_ids = np.array(clip_ids).astype(np.int64)
        if len(clip_ids) == 0:
            return 1
        max_clip = np.amax(clip_ids)
        return max_clip + 1

    def add_track(
        self,
        clip_id,
        track,
        cropped_data,
        sample_frames=None,
        opts=None,
        start_time=None,
        end_time=None,
        prediction=None,
        prediction_classes=None,
        original_thermal=None,
    ):
        """
        Adds track to database.
        :param clip_id: id of the clip to add track to write
        :param cropped_data: data for track, list of numpy arrays of shape [channels, height, width]
        :param track: the original track record, used to get stats for track
        :param opts: additional parameters used when creating dataset, if not provided defaults to no compression.
        """

        track_id = str(track.get_id())
        logging.info("Adding track %s", track_id)
        if opts is None:
            opts = {}
        with HDF5Manager(self.database, "a") as f:
            clips = f["clips"]
            clip_node = clips[clip_id]
            has_prediction = False
            track_node = clip_node.create_group(track_id)
            cropped_frame = track_node.create_group("cropped")
            if "original_frames" in clip_node:
                original_group = clip_node["original_frames"]
            else:
                original_group = clip_node.create_group("original_frames")
            skipped_frames = []
            # write each frame out individually, as they will probably be different sizes.
            original = None
            for frame_i, cropped in enumerate(cropped_data):
                if original_thermal is not None:
                    original = original_thermal[frame_i]
                # using a chunk size of 1 for channels has the advantage that we can quickly load just one channel

                if (
                    original is not None
                    and str(frame_i + track.start_frame) not in original_group
                ):
                    thermal_node = original_group.create_dataset(
                        str(frame_i + track.start_frame),
                        original.shape,
                        chunks=original.shape,
                        **opts,
                        dtype=np.int16,
                    )
                    thermal_node[:, :] = original
                # if original is not None:
                if cropped.thermal.size > 0:
                    height, width = cropped.shape
                    chunks = (1, height, width)
                    cropped_array = cropped.as_array()
                    dims = cropped_array.shape
                    frame_node = cropped_frame.create_dataset(
                        str(frame_i), dims, chunks=chunks, **opts, dtype=np.int16
                    )
                    frame_node[:, :, :] = cropped_array
                else:
                    skipped_frames.append(frame_i + track.start_frame)

            # write out attributes
            track_stats = track.get_stats()
            node_attrs = track_node.attrs
            node_attrs["id"] = track_id
            if track.tags:
                node_attrs["track_tags"] = json.dumps(track.tags)
            if sample_frames is not None:
                node_attrs["sample_frames"] = np.uint16(sample_frames)
            node_attrs["tag"] = track.tag
            node_attrs["frames"] = len(cropped_data)
            node_attrs["skipped_frames"] = np.uint16(skipped_frames)
            node_attrs["start_frame"] = track.start_frame
            node_attrs["end_frame"] = track.end_frame
            if track.predictions is not None:
                self.add_prediction_data(
                    track_node,
                    track.predictions,
                )
                has_prediction = True
            if track.confidence:
                node_attrs["confidence"] = track.confidence
            if start_time:
                node_attrs["start_time"] = start_time.isoformat()
            if end_time:
                node_attrs["end_time"] = end_time.isoformat()

            for name, value in track_stats._asdict().items():
                node_attrs[name] = value
            # frame history
            node_attrs["mass_history"] = np.int32(
                [bounds.mass for bounds in track.bounds_history]
            )
            node_attrs["bounds_history"] = np.int16(
                [bounds.to_array() for bounds in track.bounds_history]
            )
            f.flush()

            # mark the record as have been writen to.
            # this means if we are interupted part way through the track will be overwritten
            clip_node.attrs["has_prediction"] = has_prediction


def hdf5_attributes_dictionary(dataset):
    result = {}
    for key, value in dataset.attrs.items():
        result[key] = value
    return result



SOCKET_NAME = "/var/run/classifier"
import json
import sys
import socket
import argparse
import logging
from ml_tools.logs import init_logging
from config.config import Config
import time


def str2bool(v):
    if isinstance(v, bool):
        return v
    if v.lower() in ("yes", "true", "t", "y", "1"):
        return True
    elif v.lower() in ("no", "false", "f", "n", "0"):
        return False
    else:
        raise argparse.ArgumentTypeError("Boolean value expected.")


def parse_args(cmd_args=None):
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--cache",
        type=str2bool,
        nargs="?",
        const=True,
        default=None,
        help="Dont keep video frames in memory for classification later, but cache them to disk (Best for large videos, but slower)",
    )
    parser.add_argument(
        "-o",
        "--meta-to-stdout",
        action="count",
        help="Print metadata to stdout instead of saving to file.",
    )
    parser.add_argument(
        "source",
        help="a CPTV file to classify",
    )
    if cmd_args is None:
        args = parser.parse_args()
    else:
        args = parser.parse_args(cmd_args)
    return args


def main(cmd_args=None):
    args = parse_args(cmd_args)
    init_logging()
    logging.info("Classifying %s", args.source)
    sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
    config = Config.load_from_file()
    results = ""
    try:
        count = 0
        retries = 1
        while True:
            try:
                sock.connect(config.classify.service_socket)
                logging.info("Connected to %s ", config.classify.service_socket)
                break
            except Exception as ex:
                if count >= retries:
                    raise ex
                count += 1
                logging.warning(
                    "Could not connect to %s retrying in 10s error was %s",
                    config.classify.service_socket,
                    ex,
                )
                time.sleep(10)
        data = {"file": args.source, "cache": args.cache, "reuse_frames": False}
        sock.send(json.dumps(data).encode())

        results = read_all(sock).decode()
        meta_data = json.loads(str(results))
        if "error" in meta_data:
            raise Exception(meta_data["error"])

            logging.error("Error classifying %s %s", args.source, meta_data["error"])
    except socket.error as msg:
        print(msg)
        sys.exit(1)
    finally:
        # Clean up the connection
        sock.close()
    if args.meta_to_stdout:
        print(str(results))
    return str(results)


def read_all(socket):
    size = 4096
    data = bytearray()

    while True:
        packet = socket.recv(size)
        if packet:
            data.extend(packet)
        else:
            break
    return data


if __name__ == "__main__":
    main()



# Copyright 2020 Google Research. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from pathlib import Path
from multiprocessing import Process, Queue
import os
from absl import logging
import numpy as np
import tensorflow as tf


def process_job(queue, labels, base_dir, save_data, writer_i, extra_args):
    import gc

    pid = os.getpid()

    name = f"{writer_i}-{pid}.tfrecord"
    logging.info("Writing to %s", name)
    options = tf.io.TFRecordOptions(compression_type="GZIP")
    writer = tf.io.TFRecordWriter(str(base_dir / name), options=options)
    i = 0
    saved = 0
    files = 0
    num_frames = extra_args.get("num_frames", 25)
    while True:
        i += 1
        samples = queue.get()
        try:
            if samples == "DONE":
                writer.close()
                break
            else:
                if len(samples) == 0:
                    continue
                saved += save_data(samples, writer, labels, extra_args)
                files += 1
                del samples

                if i % int(2500 / num_frames) == 0:
                    logging.info("Saved %s ", files)
                    gc.collect()
                    writer.flush()
        except:
            logging.error("Process_job error %s", samples[0].source_file, exc_info=True)


def create_tf_records(
    dataset,
    output_path,
    labels,
    save_data,
    num_shards=1,
    augment=False,
    **extra_args,
):
    output_path = Path(output_path)
    if output_path.is_dir():
        logging.info("Clearing dir %s", output_path)
        for child in output_path.glob("*"):
            if child.is_file():
                child.unlink()
    output_path.mkdir(parents=True, exist_ok=True)
    samples_by_source = dataset.get_samples_by_source()
    source_files = list(samples_by_source.keys())
    np.random.shuffle(source_files)
    num_labels = len(dataset.labels)
    logging.info(
        "writing to output path: %s for %s samples", output_path, len(samples_by_source)
    )
    num_processes = 8
    writer_i = 0
    index = 0
    jobs_per_process = 600 * num_processes
    try:
        while index < len(source_files):
            job_queue = Queue()
            processes = []
            for i in range(num_processes):
                p = Process(
                    target=process_job,
                    args=(
                        job_queue,
                        labels,
                        output_path,
                        save_data,
                        writer_i,
                        extra_args,
                    ),
                )
                processes.append(p)
                p.start()
                added = 0
            writer_i += 1
            for source_file in source_files[index : index + jobs_per_process]:
                job_queue.put((samples_by_source[source_file]))
                added += 1

            index += jobs_per_process
            logging.info("Processing %d", job_queue.qsize())
            for i in range(len(processes)):
                job_queue.put(("DONE"))
            for process in processes:
                try:
                    process.join()
                except KeyboardInterrupt:
                    logging.info("KeyboardInterrupt, terminating.")
                    for process in processes:
                        process.terminate()
                    exit()
            logging.info("Saved %s", len(dataset.samples_by_id))

    except:
        logging.error("Error saving track info", exc_info=True)



from abc import ABC, abstractmethod


class Tracker(ABC):
    @abstractmethod
    def add_region(self, region):
        """Add region"""

    @property
    @abstractmethod
    def last_bound(self):
        """Tracker version"""
        ...

    @property
    @abstractmethod
    def tracker_version(self):
        """Tracker version"""
        ...

    @property
    @abstractmethod
    def frames_since_target_seen(self):
        """frames_since_target_seen version"""
        ...

    @property
    @abstractmethod
    def blank_frames(self):
        """blank_frames version"""
        ...

    @abstractmethod
    def predicted_velocity(self):
        """predicted_velocity version"""
        ...

    @property
    @abstractmethod
    def tracking(self):
        """tracking"""
        ...



"""
classifier-pipeline - this is a server side component that manipulates cptv
files and to create a classification model of animals present
Copyright (C) 2018, The Cacophony Project

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
"""

import logging
import math
import numpy as np
from collections import namedtuple

from ml_tools.rectangle import Rectangle
from track.region import Region
from .kalman import Kalman
from ml_tools.tools import eucl_distance_sq
from ml_tools.datasetstructures import get_segments, SegmentHeader, SegmentType
import logging
from track.tracker import Tracker


class RegionTracker(Tracker):
    # number of frames required before using kalman estimation
    MIN_KALMAN_FRAMES = 18

    # THERMAL VALUES
    # GP Need to put in config per camera type
    # MAX_DISTANCE = 2000
    #
    # TRACKER_VERSION = 1
    # BASE_DISTANCE_CHANGE = 450
    # # minimum region mass change
    # MIN_MASS_CHANGE = 20
    # # enforce mass growth after X seconds
    # RESTRICT_MASS_AFTER = 1.5
    # # amount region mass can change
    MASS_CHANGE_PERCENT = 0.55

    # IR VALUES
    BASE_DISTANCE_CHANGE = 11250

    # minimum region mass change
    MIN_MASS_CHANGE = 20 * 4
    # enforce mass growth after X seconds
    RESTRICT_MASS_AFTER = 1.5
    # amount region mass can change

    # MAX_DISTANCE = 2000
    MAX_DISTANCE = 30752
    BASE_VELOCITY = 8
    VELOCITY_MULTIPLIER = 10

    def __init__(self, id, tracking_config, crop_rectangle=None):
        self.track_id = id
        self.clear_run = 0
        self.kalman_tracker = Kalman()
        self._frames_since_target_seen = 0
        self.frames = 0
        self._blank_frames = 0
        self._last_bound = None
        self.crop_rectangle = crop_rectangle
        self._tracking = False
        self.type = tracking_config.type
        self.min_mass_change = tracking_config.params.get(
            "min_mass_change", RegionTracker.MIN_MASS_CHANGE
        )
        self.max_distance = tracking_config.params.get(
            "max_distance", RegionTracker.MAX_DISTANCE
        )
        self.base_distance_change = tracking_config.params.get(
            "base_distance_change", RegionTracker.BASE_DISTANCE_CHANGE
        )
        self.restrict_mass_after = tracking_config.params.get(
            "restrict_mass_after", RegionTracker.RESTRICT_MASS_AFTER
        )
        self.mass_change_percent = tracking_config.params.get(
            "mass_change_percent", RegionTracker.MASS_CHANGE_PERCENT
        )
        self.velocity_multiplier = tracking_config.params.get(
            "velocity_multiplier", RegionTracker.VELOCITY_MULTIPLIER
        )
        self.base_velocity = tracking_config.params.get(
            "base_velocity", RegionTracker.BASE_VELOCITY
        )
        self.max_blanks = tracking_config.params.get("max_blanks", 18)

    @property
    def tracking(self):
        return self._tracking

    @property
    def last_bound(self):
        return self._last_bound

    def get_size_change(self, current_area, region: Region):
        """
        Gets a value representing the difference in regions sizes
        """

        # ratio of 1.0 = 20 points, ratio of 2.0 = 10 points, ratio of 3.0 = 0 points.
        # area is padded with 50 pixels so small regions don't change too much
        size_difference = abs(region.area - current_area) / (current_area + 50)

        return size_difference

    def match(self, regions, track):
        scores = []
        avg_mass = track.average_mass()
        max_distances = self.get_max_distance_change(track)
        for region in regions:
            size_change = self.get_size_change(track.average_area(), region)
            distances = self.last_bound.average_distance(region)

            max_size_change = get_max_size_change(track, region)
            max_mass_change = self.get_max_mass_change_percent(track, avg_mass)

            logging.debug(
                "Track %s %s has max size change %s, distances %s to region %s size change %s max distance %s",
                track,
                track.last_bound,
                max_size_change,
                distances,
                region,
                size_change,
                max_distances,
            )
            # only for thermal
            if type == "thermal":
                # GP should figure out good values for the 3 distances rather than the mean
                distances = [np.mean(distances)]
                max_distances = max_distances[:1]
            else:
                distances = [(distances[0] + distances[2]) / 2]
                max_distances = max_distances[:1]

            if max_mass_change and abs(avg_mass - region.mass) > max_mass_change:
                logging.debug(
                    "track %s region mass %s deviates too much from %s for region %s",
                    track.get_id(),
                    region.mass,
                    avg_mass,
                    region,
                )
                continue
            skip = False
            for distance, max_distance in zip(distances, max_distances):
                if max_distance is None:
                    continue
                if distance > max_distance:
                    logging.debug(
                        "track %s distance score %s bigger than max distance %s for region %s",
                        track.get_id(),
                        distance,
                        max_distance,
                        region,
                    )
                    skip = True
                    break
                    # continue
            if skip:
                continue

            if size_change > max_size_change:
                logging.debug(
                    "track % size_change %s bigger than max size_change %s for region %s",
                    track.get_id(),
                    size_change,
                    max_size_change,
                    region,
                )
                continue
            # only for thermal
            if type == "ir":
                distance_score = np.mean(distances)
            else:
                # GP should figure out good values for the 3 distances rather than the mean
                distance_score = distances[0]

            scores.append((distance_score, track, region))
        return scores

    def add_region(self, region):
        self.frames += 1
        if region.blank:
            self._blank_frames += 1
            self._frames_since_target_seen += 1
            stop_tracking = min(
                2 * (self.frames - self._frames_since_target_seen),
                self.max_blanks,
            )
            self._tracking = self._frames_since_target_seen < stop_tracking
        else:
            if self._frames_since_target_seen != 0:
                self.clear_run = 0
            self.clear_run += 1
            self._tracking = True
            self.kalman_tracker.correct(region)
            self._frames_since_target_seen = 0

        prediction = self.kalman_tracker.predict()
        self.predicted_mid = (prediction[0][0], prediction[1][0])
        self._last_bound = region

    @property
    def blank_frames(self):
        return self._blank_frames

    @property
    def frames_since_target_seen(self):
        return self._frames_since_target_seen

    @property
    def nonblank_frames(self):
        return self.frames - self._blank_frames

    def predicted_velocity(self):
        if (
            self.last_bound is None
            or self.nonblank_frames <= RegionTracker.MIN_KALMAN_FRAMES
        ):
            return (0, 0)
        pred_vel_x = self.predicted_mid[0] - self.last_bound.centroid[0]
        pred_vel_y = self.predicted_mid[1] - self.last_bound.centroid[1]

        return (pred_vel_x, pred_vel_y)

    def add_blank_frame(self):
        kalman_amount = (
            self.frames
            - RegionTracker.MIN_KALMAN_FRAMES
            - self._frames_since_target_seen * 2
        )

        if kalman_amount > 0:
            region = Region(
                int(self.predicted_mid[0] - self.last_bound.width / 2.0),
                int(self.predicted_mid[1] - self.last_bound.height / 2.0),
                self.last_bound.width,
                self.last_bound.height,
                centroid=[self.predicted_mid[0], self.predicted_mid[1]],
            )
            if self.crop_rectangle:
                region.crop(self.crop_rectangle)
        else:
            region = self.last_bound.copy()
        region.blank = True
        region.mass = 0
        region.pixel_variance = 0
        region.frame_number = self.last_bound.frame_number + 1

        self.add_region(region)
        return region

    def tracker_version(self):
        return f"RegionTracker-{RegionTracker.TRACKER_VERSION}"

    def get_max_distance_change(self, track):
        x, y = track.velocity
        # x = max(x, 2)
        # y = max(y, 2)
        if len(track) == 1:
            x = self.base_velocity
            y = self.base_velocity
        x = self.velocity_multiplier * x
        y = self.velocity_multiplier * y
        velocity_distance = x * x + y * y

        pred_vel = track.predicted_velocity()
        logging.debug(
            "%s velo %s pred vel %s vel distance %s",
            track,
            track.velocity,
            track.predicted_velocity(),
            velocity_distance,
        )
        pred_distance = pred_vel[0] * pred_vel[0] + pred_vel[1] * pred_vel[1]
        pred_distance = max(velocity_distance, pred_distance)
        max_distance = self.base_distance_change + max(velocity_distance, pred_distance)
        # max top left, max between predicted and region, max between right bottom
        distances = [max_distance, None, max_distance]
        return distances

    def get_max_mass_change_percent(self, track, average_mass):
        if self.mass_change_percent is None:
            return None
        if len(track) > self.restrict_mass_after * track.fps:
            vel = track.velocity
            mass_percent = self.mass_change_percent
            if np.sum(np.abs(vel)) > 5:
                # faster tracks can be a bit more deviant
                mass_percent = mass_percent + 0.1
            return max(
                self.min_mass_change,
                average_mass * mass_percent,
            )
        else:
            return None


def get_max_size_change(track, region):
    exiting = region.is_along_border and not track.last_bound.is_along_border
    entering = not exiting and track.last_bound.is_along_border
    region_percent = 1.5
    if len(track) < 5:
        # may increase at first
        region_percent = 2
    vel = np.sum(np.abs(track.velocity))
    if entering or exiting:
        region_percent = 2
        if vel > 10:
            region_percent *= 3
    elif vel > 10:
        region_percent *= 2
    return region_percent


class ThumbInfo:
    def __init__(self, track_id):
        self.points = -1
        self.region = None
        self.thumb = None
        self.thumb_frame = None
        self.last_frame_check = None
        self.predicted_tag = None
        self.predicted_confidence = None
        self.track_id = track_id

    # score thumbs based on not being false positive having priority
    # then if sure of the prediction (above 80%) choose the most confidence
    # and then choose the one with more points
    def score(self):
        confidence_threshold = 80

        score = self.points
        score_offset = 100000

        if self.predicted_tag is not None:
            if self.predicted_tag != "false-positive":
                score = score + 1000 * score_offset
                if self.predicted_confidence > confidence_threshold:
                    confidence = self.predicted_confidence
                else:
                    confidence = 0
            else:
                confidence = 100 - self.predicted_confidence

            score = score + confidence * score_offset
        # logging.info("%s score for %s %s %s is %s",self.track_id,self.points,self.predicted_tag,self.predicted_confidence, score)
        return score


class Track:
    """Bounds of a tracked object over time."""

    # keeps track of which id number we are up to.
    _track_id = 1

    # Percentage increase that is considered jitter, e.g. if a region gets
    # 30% bigger or smaller
    JITTER_THRESHOLD = 0.3
    # must change atleast 5 pixels to be considered for jitter
    MIN_JITTER_CHANGE = 5

    def __init__(
        self,
        clip_id,
        id=None,
        fps=9,
        tracking_config=None,
        crop_rectangle=None,
        tracker_version=None,
    ):
        """
        Creates a new Track.
        :param id: id number for track, if not specified is provided by an auto-incrementer
        """
        self.in_trap = False
        self.trap_reported = False
        self.trigger_frame = None
        self.direction = 0
        self.trap_tag = None
        if not id:
            self._id = Track._track_id
            Track._track_id += 1
        else:
            self._id = id
        self.clip_id = clip_id
        self.start_frame = None
        self.start_s = None
        self.end_s = None
        self.fps = fps
        self.current_frame_num = None
        self.frame_list = []
        # our bounds over time
        self.bounds_history = []
        # number frames since we lost target.

        self.vel_x = []
        self.vel_y = []
        # the tag for this track
        self.tag = "unknown"
        self.prev_frame_num = None
        self.confidence = None
        self.max_novelty = None
        self.avg_novelty = None

        self.from_metadata = False
        self.tags = None
        self.predictions = None
        self.predicted_class = None
        self.predicted_confidence = None

        self.all_class_confidences = None
        self.prediction_classes = None

        self.crop_rectangle = crop_rectangle

        self.predictions = None
        self.predicted_tag = None
        self.predicted_confidence = None

        self.all_class_confidences = None
        self.prediction_classes = None
        self.tracker_version = tracker_version

        self.tracker = None
        if tracking_config is not None:
            self.tracker = self.get_tracker(tracking_config)
        # self.tracker = RegionTracker(
        #     self.get_id(), tracking_config, self.crop_rectangle
        #
        self.thumb_info = None
        self.score = None

    def get_tracker(self, tracking_config):
        tracker = tracking_config.tracker
        if tracker == "RegionTracker":
            return RegionTracker(self.get_id(), tracking_config, self.crop_rectangle)
        else:
            raise Exception(f"Cant find for tracker {tracker}")

    @property
    def blank_frames(self):
        if self.tracker is None:
            return 0
        return self.tracker.blank_frames

    @property
    def tracking(self):
        return self.tracker.tracking

    @property
    def frames_since_target_seen(self):
        return self.tracker.frames_since_target_seen

    def match(self, regions):
        return self.tracker.match(regions, self)

    def get_segments(
        self,
        # frame_temp_median,
        segment_width,
        segment_frame_spacing=9,
        repeats=1,
        min_frames=0,
        segment_frames=None,
        segment_types=[SegmentType.ALL_RANDOM],
        from_last=None,
        max_segments=None,
        ffc_frames=None,
        dont_filter=False,
        filter_by_fp=False,
        min_segments=1,
    ):
        if from_last is not None:
            if from_last == 0:
                return []
            regions = np.array(self.bounds_history[-from_last:])
            start_frame = regions[0].frame_number
        else:
            start_frame = self.start_frame
            regions = np.array(self.bounds_history)

        # frame_temp_median = np.uint16(frame_temp_median)
        segments = []
        if segment_frames is not None:
            mass_history = np.uint16([region.mass for region in regions])
            for frames in segment_frames:
                relative_frames = frames - self.start_frame
                mass_slice = mass_history[relative_frames]
                segment_mass = np.sum(mass_slice)
                segment = SegmentHeader(
                    self.clip_id,
                    self._id,
                    start_frame=start_frame,
                    frames=len(frames),
                    weight=1,
                    mass=segment_mass,
                    label=None,
                    regions=regions[relative_frames],
                    # frame_temp_median=frame_temp_median[relative_frames],
                    frame_indices=frames,
                )
                segments.append(segment)
        else:
            segments, _ = get_segments(
                self.clip_id,
                self._id,
                start_frame,
                segment_frame_spacing=segment_frame_spacing,
                segment_width=segment_width,
                regions=regions,
                ffc_frames=ffc_frames,
                repeats=repeats,
                min_frames=min_frames,
                segment_types=segment_types,
                max_segments=max_segments,
                dont_filter=dont_filter,
                min_segments=min_segments,
            )

        return segments

    @classmethod
    def from_region(cls, clip, region, tracker_version=None, tracking_config=None):
        track = cls(
            clip.get_id(),
            fps=clip.frames_per_second,
            tracker_version=tracker_version,
            crop_rectangle=clip.crop_rectangle,
            tracking_config=tracking_config,
        )
        track.start_frame = region.frame_number
        track.start_s = region.frame_number / float(clip.frames_per_second)
        track.add_region(region)
        return track

    def get_id(self):
        return self._id

    def add_prediction_info(self, track_prediction):
        logging.warn("TODO add prediction info needs to be implemented")
        return

    def load_track_meta(
        self,
        track_meta,
        frames_per_second,
        tag_precedence=None,
        min_confidence=0.8,
    ):
        self.tracker_version = track_meta.get("tracker_version", "unknown")
        self.from_metadata = True
        self._id = track_meta["id"]
        extra_info = track_meta
        if "data" in track_meta:
            extra_info = track_meta["data"]
        if "start_s" in extra_info:
            self.start_s = extra_info["start_s"]
            self.end_s = extra_info["end_s"]
        else:
            self.start_s = extra_info["start"]
            self.end_s = extra_info["end"]
        self.fps = frames_per_second

        self.tags = track_meta.get("tags")
        tag = Track.get_best_human_tag(self.tags, tag_precedence, min_confidence)
        if tag:
            self.tag = tag["what"]
            self.confidence = tag["confidence"]

        self.stats = TrackMovementStatistics(score=track_meta.get("tracking_score", 0))

        positions = track_meta.get("positions")
        if not positions:
            return False
        self.bounds_history = []
        self.frame_list = []
        for i, position in enumerate(positions):
            if isinstance(position, list):
                region = Region.region_from_array(position[1])
                if region.frame_number is None:
                    frame_number = round(position[0] * frames_per_second)

                    region.frame_number = frame_number
            else:
                region = Region.region_from_json(position)
                if region.frame_number is None:
                    if "frameTime" in position:
                        if i == 0:
                            region.frame_number = position["frameTime"] * 9
                        else:
                            region.frame_number = (
                                self.bounds_history[0].frame_number + i
                            )
                    else:
                        raise Exception("No frame number info for track")
            if self.start_frame is None:
                self.start_frame = region.frame_number
            # self.end_frame = region.frame_number
            self.bounds_history.append(region)
            self.frame_list.append(region.frame_number)
        self.current_frame_num = 0
        return True

    #
    # def add_frame(self, frame):
    #     if self.stable:
    #         ok, bbox = self.tracker.update(frame.thermal)
    #         if ok:
    #             r = Region.from_ltwh(bbox[0], bbox[1], bbox[2], bbox[3])
    #
    #             r.crop(self.crop_rectangle)
    #
    #             sub_filtered = r.subimage(frame.filtered)
    #             r.calculate_mass(sub_filtered, 1)
    #             r.frame_number = frame.frame_number
    #             self.add_region(r, frame)
    #
    #         else:
    #             self.add_blank_frame()
    #
    def add_region(self, region):
        if self.prev_frame_num and region.frame_number:
            frame_diff = region.frame_number - self.prev_frame_num - 1
            for _ in range(frame_diff):
                self.add_blank_frame()
        self.tracker.add_region(region)
        self.bounds_history.append(region)
        # self.end_frame = region.frame_number
        self.prev_frame_num = region.frame_number
        self.update_velocity()

    def update_velocity(self):
        if len(self.bounds_history) >= 2:
            self.vel_x.append(
                self.bounds_history[-1].centroid[0]
                - self.bounds_history[-2].centroid[0]
            )
            self.vel_y.append(
                self.bounds_history[-1].centroid[1]
                - self.bounds_history[-2].centroid[1]
            )
        else:
            self.vel_x.append(0)
            self.vel_y.append(0)

    def crop_regions(self):
        if self.crop_rectangle is None:
            logging.info("No crop rectangle to crop with")
            return
        for region in self.bounds_history:
            region.crop(self.crop_rectangle)

    def add_frame_for_existing_region(self, frame, mass_delta_threshold, prev_filtered):
        region = self.bounds_history[self.current_frame_num]
        if prev_filtered is not None:
            prev_filtered = region.subimage(prev_filtered)
        filtered = region.subimage(frame.filtered)
        region.calculate_mass(filtered, mass_delta_threshold)
        region.calculate_variance(filtered, prev_filtered)
        if self.prev_frame_num and frame.frame_number:
            frame_diff = frame.frame_number - self.prev_frame_num - 1
            for _ in range(frame_diff):
                self.add_blank_frame()
        if self.tracker:
            self.tracker.add_region(region)

        self.update_velocity()
        self.prev_frame_num = frame.frame_number
        self.current_frame_num += 1

    def average_area(self):
        """Average mass of last 5 frames that weren't blank"""
        avg_area = 0
        count = 0
        for i in range(len(self.bounds_history)):
            bound = self.bounds_history[-i - 1]
            if not bound.blank:
                avg_area += bound.area
                count += 1
            if count == 5:
                break
        if count == 0:
            return 0
        return avg_area / count

    def average_mass(self):
        """Average mass of last 5 frames that weren't blank"""
        avg_mass = 0
        count = 0
        for i in range(len(self.bounds_history)):
            bound = self.bounds_history[-i - 1]
            if not bound.blank:
                avg_mass += bound.mass
                count += 1
            if count == 5:
                break
        if count == 0:
            return 0
        return avg_mass / count

    def add_blank_frame(self):
        """Maintains same bounds as previously, does not reset framce_since_target_seen counter"""
        if self.tracker:
            region = self.tracker.add_blank_frame()
        self.bounds_history.append(region)
        self.prev_frame_num = region.frame_number
        self.update_velocity()

    def calculate_stats(self):
        """
        Calculates statistics for this track, including how much it moves, and a score indicating how likely it is
        that this is a good track. This should be done once tracking is finished
        # :return: a TrackMovementStatistics record
        """

        if len(self) <= 1:
            self.stats = TrackMovementStatistics()
            return
        # get movement vectors only from non blank regions
        non_blank = [bound for bound in self.bounds_history if not bound.blank]
        mass_history = [int(bound.mass) for bound in non_blank]
        variance_history = [
            bound.pixel_variance for bound in non_blank if bound.pixel_variance
        ]
        movement = 0
        max_offset = 0

        frames_moved = 0
        avg_vel = 0
        first_point = self.bounds_history[0].mid
        for i, (vx, vy) in enumerate(zip(self.vel_x, self.vel_y)):
            region = self.bounds_history[i]
            if not region.blank:
                avg_vel += abs(vx) + abs(vy)
            if i == 0:
                continue

            if region.blank or self.bounds_history[i - 1].blank:
                continue
            if region.has_moved(self.bounds_history[i - 1]) or region.is_along_border:
                distance = (vx**2 + vy**2) ** 0.5
                movement += distance
                offset = eucl_distance_sq(first_point, region.mid)
                max_offset = max(max_offset, offset)
                frames_moved += 1
        avg_vel = avg_vel / len(mass_history)
        # the standard deviation is calculated by averaging the per frame variances.
        # this ends up being slightly different as I'm using /n rather than /(n-1) but that
        # shouldn't make a big difference as n = width*height*frames which is large.
        max_offset = math.sqrt(max_offset)
        delta_std = float(np.mean(variance_history)) ** 0.5
        jitter_bigger = 0
        jitter_smaller = 0
        for i, bound in enumerate(self.bounds_history[1:]):
            prev_bound = self.bounds_history[i]
            if prev_bound.is_along_border or bound.is_along_border:
                continue
            height_diff = bound.height - prev_bound.height
            width_diff = prev_bound.width - bound.width
            thresh_h = max(
                Track.MIN_JITTER_CHANGE, prev_bound.height * Track.JITTER_THRESHOLD
            )
            thresh_v = max(
                Track.MIN_JITTER_CHANGE, prev_bound.width * Track.JITTER_THRESHOLD
            )
            if abs(height_diff) > thresh_h:
                if height_diff > 0:
                    jitter_bigger += 1
                else:
                    jitter_smaller += 1
            elif abs(width_diff) > thresh_v:
                if width_diff > 0:
                    jitter_bigger += 1
                else:
                    jitter_smaller += 1

        movement_points = (movement**0.5) + max_offset
        delta_points = delta_std * 25.0
        jitter_percent = int(
            round(100 * (jitter_bigger + jitter_smaller) / float(self.frames))
        )

        blank_percent = int(round(100.0 * self.blank_frames / self.frames))
        score = (
            min(movement_points, 100)
            + min(delta_points, 100)
            + (100 - jitter_percent)
            + (100 - blank_percent)
        )
        stats = TrackMovementStatistics(
            movement=float(movement),
            max_offset=float(max_offset),
            average_mass=float(np.mean(mass_history)),
            median_mass=float(np.median(mass_history)),
            delta_std=float(delta_std),
            score=float(score),
            region_jitter=jitter_percent,
            jitter_bigger=jitter_bigger,
            jitter_smaller=jitter_smaller,
            blank_percent=blank_percent,
            frames_moved=frames_moved,
            mass_std=float(np.std(mass_history)),
            average_velocity=float(avg_vel),
        )
        self.stats = stats

    def smooth(self, frame_bounds: Rectangle):
        """
        Smooths out any quick changes in track dimensions
        :param frame_bounds The boundaries of the video frame.
        """
        if len(self.bounds_history) == 0:
            return

        new_bounds_history = []
        prev_frame = self.bounds_history[0]
        current_frame = self.bounds_history[0]
        next_frame = self.bounds_history[1]

        for i in range(len(self.bounds_history)):
            prev_frame = self.bounds_history[max(0, i - 1)]
            current_frame = self.bounds_history[i]
            next_frame = self.bounds_history[min(len(self.bounds_history) - 1, i + 1)]

            frame_x = current_frame.centroid[0]
            frame_y = current_frame.centroid[1]
            frame_width = (
                prev_frame.width + current_frame.width + next_frame.width
            ) / 3
            frame_height = (
                prev_frame.height + current_frame.height + next_frame.height
            ) / 3
            frame = Region(
                int(frame_x - frame_width / 2),
                int(frame_y - frame_height / 2),
                int(frame_width),
                int(frame_height),
            )
            frame.crop(frame_bounds)

            new_bounds_history.append(frame)

        self.bounds_history = new_bounds_history

    def trim(self):
        """
        Removes empty frames from start and end of track
        """
        mass_history = [int(bound.mass) for bound in self.bounds_history]
        median_mass = np.median(mass_history)
        filter_mass = 0.005 * median_mass
        filter_mass = max(filter_mass, 2)
        start = 0
        logging.debug(
            "Triming track with median % and filter mass %s", median_mass, filter_mass
        )
        while start < len(self) and mass_history[start] <= filter_mass:
            start += 1
        end = len(self) - 1

        while end > 0 and mass_history[end] <= filter_mass:
            if self.tracker and self.frames_since_target_seen > 0:
                self.tracker._frames_since_target_seen -= 1
                self.tracker._blank_frames -= 1
            end -= 1
        if end < start:
            self.bounds_history = []
            self.vel_x = []
            self.vel_y = []
            if self.tracker:
                self.tracker._blank_frames = 0
        else:
            self.start_frame += start
            self.bounds_history = self.bounds_history[start : end + 1]
            self.vel_x = self.vel_x[start : end + 1]
            self.vel_y = self.vel_y[start : end + 1]
        self.start_s = self.start_frame / float(self.fps)

    def get_overlap_ratio(self, other_track, threshold=0.05):
        """
        Checks what ratio of the time these two tracks overlap.
        :param other_track: the other track to compare with
        :param threshold: how much frames must be overlapping to be counted
        :return: the ratio of frames that overlap
        """

        if len(self) == 0 or len(other_track) == 0:
            return 0.0

        start = max(self.start_frame, other_track.start_frame)
        end = min(self.end_frame, other_track.end_frame)

        frames_overlapped = 0

        for pos in range(start, end + 1):
            our_index = pos - self.start_frame
            other_index = pos - other_track.start_frame
            if (
                our_index >= 0
                and other_index >= 0
                and our_index < len(self)
                and other_index < len(other_track)
            ):
                our_bounds = self.bounds_history[our_index]
                if our_bounds.area == 0:
                    continue
                other_bounds = other_track.bounds_history[other_index]
                overlap = our_bounds.overlap_area(other_bounds) / our_bounds.area
                if overlap >= threshold:
                    frames_overlapped += 1

        return frames_overlapped / len(self)

    def set_end_s(self, fps):
        if len(self) == 0:
            self.end_s = self.start_s
            return
        self.end_s = (self.end_frame + 1) / fps

    def predicted_velocity(self):
        return self.tracker.predicted_velocity()

    def update_trapped_state(self):
        if self.in_trap:
            return self.in_trap
        min_frames = 2
        if len(self.bounds_history) < min_frames:
            return False
        self.in_trap = all(r.in_trap for r in self.bounds_history[-min_frames:])
        return self.in_trap

    @property
    def end_frame(self):
        if len(self.bounds_history) == 0:
            return self.start_frame
        return self.bounds_history[-1].frame_number

    @property
    def nonblank_frames(self):
        return self.end_frame + 1 - self.start_frame - self.blank_frames

    @property
    def frames(self):
        return self.end_frame + 1 - self.start_frame

    @property
    def last_mass(self):
        return self.bounds_history[-1].mass

    @property
    def velocity(self):
        return self.vel_x[-1], self.vel_y[-1]

    @property
    def last_bound(self) -> Region:
        return self.bounds_history[-1]

    def __repr__(self):
        return "Track: {} frames# {}".format(self.get_id(), len(self))

    def __len__(self):
        return len(self.bounds_history)

    def start_and_end_in_secs(self):
        if self.end_s is None:
            if len(self) == 0:
                self.end_s = self.start_s
            else:
                self.end_s = (self.end_frame + 1) / self.fps

        return (self.start_s, self.end_s)

    def get_metadata(self, predictions_per_model=None):
        track_info = {}
        start_s, end_s = self.start_and_end_in_secs()

        track_info["id"] = self.get_id()
        if self.in_trap:
            track_info["trap_triggered"] = self.in_trap
            track_info["trigger_frame"] = self.trigger_frame
            if self.trap_tag is not None:
                track_info["trap_tag"] = self.trap_tag
        track_info["tracker_version"] = self.tracker_version
        track_info["start_s"] = round(start_s, 2)
        track_info["end_s"] = round(end_s, 2)
        track_info["num_frames"] = len(self)
        track_info["frame_start"] = self.start_frame
        track_info["frame_end"] = self.end_frame
        track_info["positions"] = self.bounds_history
        track_info["tracking_score"] = 0 if self.stats is None else self.stats.score
        prediction_info = []
        if predictions_per_model:
            for model_id, predictions in predictions_per_model.items():
                prediction = predictions.prediction_for(self.get_id())
                if prediction is None:
                    continue
                prediciont_meta = prediction.get_metadata()
                prediciont_meta["model_id"] = model_id
                prediction_info.append(prediciont_meta)
        track_info["predictions"] = prediction_info
        return track_info

    @classmethod
    def get_best_human_tag(cls, track_tags, tag_precedence, min_confidence=-1):
        """returns highest precidence non AI tag from the metadata"""
        if track_tags is None:
            return None
        track_tags = [
            tag
            for tag in track_tags
            if not tag.get("automatic", False)
            and tag.get("confidence") >= min_confidence
        ]

        if not track_tags:
            return None

        tag = None
        if tag_precedence is None:
            default_prec = 100
            tag_precedence = {}
        else:
            default_prec = tag_precedence.get("default", 100)
        best = None
        for track_tag in track_tags:
            ranking = cls.tag_ranking(track_tag, tag_precedence, default_prec)
            # if 2 track_tags have same confidence ignore both
            if tag and ranking == best and track_tag["what"] != tag["what"]:
                tag = None
            elif best is None or ranking < best:
                best = ranking
                tag = track_tag
        return tag

    @staticmethod
    def tag_ranking(track_tag, precedence, default_prec):
        """returns a ranking of tags based of what they are and confidence"""

        what = track_tag.get("what")
        confidence = 1 - track_tag.get("confidence", 0)
        prec = precedence.get(what, default_prec)
        return prec + confidence


TrackMovementStatistics = namedtuple(
    "TrackMovementStatistics",
    "movement max_offset score average_mass median_mass delta_std region_jitter jitter_smaller jitter_bigger blank_percent frames_moved mass_std, average_velocity",
)
TrackMovementStatistics.__new__.__defaults__ = (0,) * len(
    TrackMovementStatistics._fields
)



from pathlib import Path
import attr
import toml
import portalocker
import os

from .locationconfig import LocationConfig
from .timewindow import RelAbsTime, TimeWindow

CONFIG_FILENAME = "config.toml"
CONFIG_DIRS = [Path(__file__).parent.parent, Path("/etc/cacophony")]


class LockSafeConfig:
    def __init__(self, filename):
        self.lock_file = filename + ".lock"
        self.filename = filename
        self.f = None
        self.lock = portalocker.Lock(
            self.lock_file, "r", flags=portalocker.LOCK_SH, timeout=1
        )
        if not os.path.exists(self.lock_file):
            f = open(self.lock_file, "w+")
            f.close()

    def __enter__(self):
        # note: we might not have to lock when in read only mode?
        # this could improve performance
        self.lock.acquire()
        self.f = open(self.filename)
        return self.f

    def __exit__(self, exc_type, exc_val, exc_tb):
        try:
            self.f.close()
        finally:
            self.lock.release()


@attr.s
class ThrottlerConfig:
    bucket_size = attr.ib()
    activate = attr.ib()
    no_motion = attr.ib()
    max_throttling_minutes = attr.ib()

    @classmethod
    def load(cls, throttler):
        return cls(
            bucket_size=RelAbsTime(
                throttler.get("bucket-size"), default_offset=10 * 60
            ).offset_s,
            activate=throttler.get("activate", True),
            no_motion=throttler.get("no-motion", 5 * 60),
            max_throttling_minutes=throttler.get("max-throttling-minutes", 60),
        )

    def as_dict(self):
        return attr.asdict(self)


@attr.s
class CameraMotionConfig:
    temp_thresh = attr.ib()
    delta_thresh = attr.ib()
    count_thresh = attr.ib()
    frame_compare_gap = attr.ib()
    one_diff_only = attr.ib()
    trigger_frames = attr.ib()
    edge_pixels = attr.ib()
    warmer_only = attr.ib()
    dynamic_thresh = attr.ib()
    run_classifier = attr.ib(default=False)
    bluetooth_beacons = attr.ib(default=False)
    tracking_events = attr.ib(default=False)
    do_tracking = attr.ib(default=False)

    @classmethod
    def defaults_for(cls, model):
        if model == "lepton3.5":
            return cls(
                temp_thresh=28000,
                delta_thresh=150,
                count_thresh=3,
                frame_compare_gap=45,
                one_diff_only=True,
                trigger_frames=2,
                edge_pixels=1,
                warmer_only=True,
                dynamic_thresh=True,
                do_tracking=False,
            )
        else:
            return cls(
                temp_thresh=2750,
                delta_thresh=50,
                count_thresh=3,
                frame_compare_gap=45,
                one_diff_only=True,
                trigger_frames=2,
                edge_pixels=1,
                warmer_only=True,
                dynamic_thresh=True,
                do_tracking=False,
            )

    @classmethod
    def load(cls, motion, model=None):
        default = CameraMotionConfig.defaults_for(model)
        motion = cls(
            temp_thresh=motion.get("temp-thresh", default.temp_thresh),
            delta_thresh=motion.get("delta-thresh", default.delta_thresh),
            count_thresh=motion.get("count-thresh", default.count_thresh),
            frame_compare_gap=motion.get(
                "frame-compare-gap", default.frame_compare_gap
            ),
            one_diff_only=motion.get("use-one-diff-only", default.one_diff_only),
            trigger_frames=motion.get("trigger-frames", default.trigger_frames),
            edge_pixels=motion.get("edge-pixels", default.edge_pixels),
            warmer_only=motion.get("warmer-only", default.warmer_only),
            dynamic_thresh=motion.get("dynamic-thresh", default.dynamic_thresh),
            run_classifier=motion.get("run-classifier", default.run_classifier),
            bluetooth_beacons=motion.get(
                "bluetooth-beacons", default.bluetooth_beacons
            ),
            tracking_events=motion.get("tracking-events", default.tracking_events),
            do_tracking=motion.get("do-tracking", default.do_tracking),
        )
        return motion

    def as_dict(self):
        return attr.asdict(self)


@attr.s
class RecorderConfig:
    preview_secs = attr.ib()
    min_secs = attr.ib()
    max_secs = attr.ib()
    rec_window = attr.ib()
    output_dir = attr.ib()
    disable_recordings = attr.ib()
    constant_recorder = attr.ib()
    use_low_power_mode = attr.ib()
    min_disk_space_mb = attr.ib()

    @classmethod
    def load(cls, recorder, window, location_config):
        return cls(
            constant_recorder=recorder.get("constant-recorder", False),
            disable_recordings=recorder.get("disable-recordings", False),
            min_secs=recorder.get("min-secs", 5),
            max_secs=recorder.get("max-secs", 600),
            preview_secs=recorder.get("preview-secs", 5),
            rec_window=TimeWindow(
                RelAbsTime(window.get("start-recording"), default_offset=30 * 60),
                RelAbsTime(window.get("stop-recording"), default_offset=30 * 60),
                *location_config.get_lat_long(use_default=True),
                location_config.altitude,
            ),
            min_disk_space_mb=recorder.get("min-disk-space-mb", 200),
            output_dir=recorder.get("output-dir", "/var/spool/cptv"),
            use_low_power_mode=recorder.get("use-low-power-mode", False),
        )


@attr.s
class DeviceSetup:
    ir = attr.ib(default=False)
    trap_size = attr.ib(default=None)
    # S or L for small or large

    @classmethod
    def load(cls, device):
        size = device.get("trap-size", "L")
        if size is not None:
            size = size.upper()
        return cls(ir=device.get("ir", False), trap_size=size)


@attr.s
class DeviceConfig:
    device_id = attr.ib()
    name = attr.ib()

    @classmethod
    def load(cls, device):
        return cls(
            name=device.get("name"),
            device_id=device.get("id"),
        )


@attr.s
class ThermalConfig:
    motion = attr.ib()
    recorder = attr.ib()
    device = attr.ib()
    location = attr.ib()
    throttler = attr.ib()
    device_setup = attr.ib()
    config_file = attr.ib()

    @classmethod
    def load_from_file(cls, filename=None, model=None):
        if not filename:
            filename = ThermalConfig.find_config()
        with LockSafeConfig(filename) as stream:
            return cls.load_from_stream(filename, stream, model)

    @classmethod
    def load_from_stream(cls, filename, stream, model=None):
        raw = toml.load(stream)
        if raw is None:
            raw = {}

        location_config = LocationConfig.load(raw.get("location", {}))
        return cls(
            config_file=filename,
            throttler=ThrottlerConfig.load(raw.get("thermal-throttler", {})),
            motion=CameraMotionConfig.load(raw.get("thermal-motion", {}), model),
            recorder=RecorderConfig.load(
                raw.get("thermal-recorder", {}), raw.get("windows", {}), location_config
            ),
            device=DeviceConfig.load(raw.get("device", {})),
            device_setup=DeviceSetup.load(raw.get("device-setup", {})),
            location=location_config,
        )

    def validate(self):
        return True

    @staticmethod
    def find_config():
        for directory in CONFIG_DIRS:
            p = directory / CONFIG_FILENAME
            if p.is_file():
                return str(p)
        raise FileNotFoundError(
            "No configuration file found.  Looking for file named '{}' in dirs {}".format(
                CONFIG_FILENAME, CONFIG_DIRS
            )
        )



import attr
import inspect

from .defaultconfig import DefaultConfig, deep_copy_map_if_key_not_exist


@attr.s
class TrackingMotionConfig(DefaultConfig):
    camera_thresholds = attr.ib()
    dynamic_thresh = attr.ib()

    @classmethod
    def load(cls, threshold):
        if inspect.isclass(threshold):
            return threshold
        return cls(
            camera_thresholds=TrackingMotionConfig.load_camera_thresholds(
                threshold.get("camera_thresholds")
            ),
            dynamic_thresh=threshold["dynamic_thresh"],
        )

    @classmethod
    def get_defaults(cls):
        thresholds = {}
        thresholds["lepton3"] = ThresholdConfig(
            camera_model="lepton3",
            temp_thresh=2900,
            background_thresh=20,
            default=True,
            min_temp_thresh=None,
            max_temp_thresh=None,
            track_min_delta=1.0,
            track_max_delta=150,
        )
        thresholds["lepton3.5"] = ThresholdConfig(
            camera_model="lepton3.5",
            temp_thresh=28000,
            background_thresh=50,
            default=False,
            min_temp_thresh=None,
            max_temp_thresh=None,
            track_min_delta=1.0,
            track_max_delta=150,
        )
        thresholds["IR"] = ThresholdConfig(
            camera_model="IR",
            temp_thresh=None,
            background_thresh=12,
            default=False,
            min_temp_thresh=None,
            max_temp_thresh=None,
            track_min_delta=1.0,
            track_max_delta=150,
        )
        return cls(
            camera_thresholds=thresholds,
            dynamic_thresh=True,
        )

    def load_camera_thresholds(raw):
        if raw is None:
            return None
        threholds = {}
        for raw_threshold in raw.values():
            threshold = ThresholdConfig.load(raw_threshold)
            threholds[threshold.camera_model] = threshold
        return threholds

    def validate(self):
        return True

    def as_dict(self):
        return attr.asdict(self)

    def threshold_for_model(self, camera_model):
        if self.camera_thresholds is None:
            return None
        threshold = self.camera_thresholds.get(camera_model)
        if not threshold:
            for model_thresh in self.camera_thresholds.values():
                if model_thresh.default:
                    return model_thresh

            return self.camera_thresholds["default-model"]
        return threshold


@attr.s
class ThresholdConfig(DefaultConfig):
    camera_model = attr.ib()
    temp_thresh = attr.ib()
    background_thresh = attr.ib()
    default = attr.ib()
    min_temp_thresh = attr.ib()
    max_temp_thresh = attr.ib()
    track_min_delta = attr.ib()
    track_max_delta = attr.ib()

    @classmethod
    def load(cls, threshold):
        defaults = cls.get_defaults()
        deep_copy_map_if_key_not_exist(defaults.as_dict(), threshold)

        return cls(
            camera_model=threshold["camera_model"],
            temp_thresh=threshold["temp_thresh"],
            background_thresh=threshold["background_thresh"],
            default=threshold["default"],
            min_temp_thresh=threshold["min_temp_thresh"],
            max_temp_thresh=threshold["max_temp_thresh"],
            track_min_delta=threshold["track_min_delta"],
            track_max_delta=threshold["track_max_delta"],
        )

    def as_dict(self):
        return attr.asdict(self)

    @classmethod
    def get_defaults(cls):
        return cls(
            camera_model="default-model",
            temp_thresh=2900,
            background_thresh=20,
            default=False,
            min_temp_thresh=None,
            max_temp_thresh=None,
            track_min_delta=1.0,
            track_max_delta=150,
        )

    def validate(self):
        return True



import dbus
import logging
import binascii
from dbus.mainloop.glib import DBusGMainLoop

DBUS_NAME = "org.cacophony.i2c"
DBUS_PATH = "/org/cacophony/i2c"


def set_recording_state(is_recording):
    bus = dbus.SystemBus(mainloop=DBusGMainLoop())
    try:
        proxy = bus.get_object(DBUS_NAME, DBUS_PATH)
        state = agent_state(proxy)
        if state is None:
            return False
        state = int(state[0])
        if is_recording:
            state = state | 4
        else:
            state = state & ~4

        res = agent_state(proxy, state)
        if res is None:
            return False

    except:
        logging.error("set recording dbus error ", exc_info=True)
        return False
    return True


def agent_state(proxy, value=None):
    if value is None:
        payload = bytearray([7])
    else:
        payload = bytearray([7, value])

    crc = binascii.crc_hqx(payload, 0x1D0F)
    crc = crc.to_bytes(2, "big")
    payload.extend(crc)

    try:
        return proxy.Tx(0x25, dbus.ByteArray(payload), 3 if value is None else 0, 1000)
    except:
        logging.error("agent state dbus error ", exc_info=True)
    return None
